module parser;
import lexer;
import std.stdio;
import std.string;

/// Fonksiyonlar arası iletişimi sağlayan dönüş değerini belirten yapı.
enum RetType{
	none, expression
}

/// İşlem önceliği tablosuna yeniden erişmemek için öncelik değerini tuttuğumuz yapı.
struct Operator{
	Type type; /// Operatör
	size_t precedence; ///İşlem önceliği
}

/**
 Belirlediğimiz gramer kurallarına göre kodu işleyecek olan sınıfımız.
 */
class Parser : Lexer{
	private:
		Token[] tokens;
		size_t i;// Şu anki token
		static size_t[Type] op_precedences;/// İşlem Önceliği tablosu
		static this(){
			/**
				İşlem önceliği yüksek olan daha büyük sayı ile gösteriliyor.
			*/
			op_precedences = [Type.plus: 2 , Type.minus: 2, Type.times: 3, Type.divide: 3];
		}

	public:
		/// Kodlarımızı lexera gönderecek ve sonra sözdizimsel analizi başlatacak olan ana fonksiyonumuz.
		void parse(string code){
			tokens = lexit(code);/// Lexerdan gelen tokenleri tokens dizisine yükle.
 			while(i < tokens.length){/// Token olduğu sürece
				switch(tokens[i].type){
					case Type.identifier:/// Bir tanımlayıcı yani değişken gelirse
						auto name = tokens[i].value;/// Sonraki token'e geçileceği için ismini değişkenin adını saklıyoruz.
						i++;/// index değerini arttırarak sonraki tokene geçiyoruz.
						if(tokens[i].type==Type.equals){///Bir tanımlama ise
							i++;/// = karakterini geçiyoruz.
							if(getIt() == RetType.expression){
								writefln("define '%s'", name);
							}else throw new Exception("= den sonra bir ifade bekleniyordu. %s geldi".format(tokens[i].type));
						}else if(getFunction()){ /// Bir fonksiyon çağırma ise
						}else{
							throw new Exception("%s değişkeni üzerinde hiçbir işlem yapılmadı".format(name));
						}
						break;
					default:/// Switch içerisindeki hiçbir yerle eşleşemezse hata fırlatıyoruz.
						throw new Exception("Beklenmeyen tür: %s".format(tokens[i].type));
				}
				if(tokens[i].type==Type.semicolon){/// Noktalı virgül kullanımını zorunlu hale getir
					i++;
					continue;
				}else throw new Exception("; bekleniyordu %s geldi".format(tokens[i].type));
			}
		}
	private:
		RetType getIt(){
			switch(tokens[i].type){
				case Type.number:
					writefln("load %s", tokens[i].value);
					i++;
					return RetType.expression;
				case Type.string:
					writefln("load \"%s\"", tokens[i].value);
					i++;
					return RetType.expression;
				case Type._true:
					writefln("load true");
					i++;
					return RetType.expression;
				case Type._false:
					writefln("load false");
					i++;
					return RetType.expression;
				case Type.identifier:
					writefln("load_var %s", tokens[i].value);
					i++;
					getFunction();
					return RetType.expression;
				default:
					return RetType.none;
			}
		}
		
		/**
			Fonksiyon çağırma işlemlerini yakalar.
		*/
		bool getFunction(){
			if(tokens[i].type==Type.lparen) i++;/// Eğer ( tokeni var ise sonraki tokene geçer yoksa fonksiyondan çıkar.
			else return false;
			size_t parametercount; /// Parametre sayısını tutar.
			if(tokens[i].type==Type.rparen) goto end; /// Parantez kapatıldı ise döngüye girmeden direk sona atlar.
			while(i < tokens.length){
				if(getIt() == RetType.none){
					throw new Exception("Bir ifade beklenirken %s geldi".format(tokens[i].type));
				}else{
					parametercount++;
					writeln("push_param");
				}
				if(tokens[i].type == Type.comma){ /// Sonraki parametre için 'virgül' kontrolü.
					i++;
					continue;
				}else break;
			}
			if(i >= tokens.length) throw new Exception("Parantezin kapatılması bekleniyordu.");
			else if(tokens[i].type!=Type.rparen) throw new Exception("Parantezin kapatılması bekleniyordu. %s geldi".format(tokens[i].type));
		end:
			i++;
			writefln("call %s", parametercount);
			return true;
		}
}